package ru.dbotthepony.mc.otm.player

import net.minecraft.nbt.CompoundTag
import net.minecraft.world.Difficulty
import net.minecraft.world.entity.player.Player
import net.minecraft.world.food.FoodConstants
import net.minecraft.world.food.FoodData
import net.minecraft.world.food.FoodProperties
import net.minecraft.world.level.GameRules
import ru.dbotthepony.mc.otm.config.IFoodRegenerationValues
import ru.dbotthepony.mc.otm.config.PlayerConfig
import ru.dbotthepony.mc.otm.core.nbt.set
import kotlin.math.max
import kotlin.math.min
import kotlin.math.roundToInt

class MatteryFoodData(private var player: Player) : FoodData() {
	private fun add(foodLevel: Int, saturation: Float) {
		this.foodLevel = min(this.foodLevel + foodLevel, PlayerConfig.Food.HARD_FOOD_LIMIT)
		this.saturationLevel = min(this.saturationLevel + saturation, this.foodLevel + PlayerConfig.Food.OVERSATURATION_LIMIT.toFloat()).coerceAtLeast(0f)
	}

	override fun eat(foodLevelModifier: Int, saturationLevelModifier: Float) {
		add(foodLevelModifier, FoodConstants.saturationByModifier(foodLevelModifier, saturationLevelModifier))
	}

	override fun eat(foodProperties: FoodProperties) {
		add(foodProperties.nutrition(), foodProperties.saturation())
	}

	override fun needsFood(): Boolean {
		return foodLevel < PlayerConfig.Food.SOFT_FOOD_LIMIT
	}

	override fun addExhaustion(exhaustion: Float) {
		this.exhaustionLevel = min(this.exhaustionLevel + exhaustion, PlayerConfig.Food.EXHAUSTION_LIMIT.toFloat())
	}

	override fun readAdditionalSaveData(compoundTag: CompoundTag) {
		super.readAdditionalSaveData(compoundTag)

		if ("lastFoodLevel" in compoundTag)
			lastFoodLevel = compoundTag.getInt("lastFoodLevel")
	}

	override fun addAdditionalSaveData(compoundTag: CompoundTag) {
		super.addAdditionalSaveData(compoundTag)
		compoundTag["lastFoodLevel"] = lastFoodLevel
	}

	private fun tickExhaustion() {
		if (exhaustionLevel >= EXHAUSTION_PER_HUNGER_POINT) {
			var points = (exhaustionLevel / EXHAUSTION_PER_HUNGER_POINT).toInt()
			exhaustionLevel %= EXHAUSTION_PER_HUNGER_POINT

			if (saturationLevel > 0f) {
				val satisfied = min(saturationLevel.roundToInt(), points)
				points -= satisfied
				saturationLevel -= satisfied
			}

			foodLevel = max(0, foodLevel - points)
		}
	}

	private fun tickRegeneration(values: IFoodRegenerationValues): Boolean {
		if (!player.isHurt || foodLevel < values.foodLimit || values.requiresSaturation && saturationLevel <= 0f)
			return false

		while (player.isHurt && foodLevel >= values.foodLimit && (!values.requiresSaturation || saturationLevel > 0f) && ++tickTimer >= values.ticks) {
			tickTimer = 0

			val healAmount = if (values.requiresSaturation && values.regenerationSlowdown) min(saturationLevel, values.upperSlowdownBound.toFloat()) / values.upperSlowdownBound.toFloat() else 1f
			player.heal(healAmount)
			addExhaustion(healAmount * 6f)
			tickExhaustion()
		}

		return tickTimer != 0
	}

	override fun tick(player: Player) {
		this.player = player

		if (player.matteryPlayer.isAndroid)
			return

		lastFoodLevel = foodLevel
		tickExhaustion()

		if (
			player.level().gameRules.getBoolean(GameRules.RULE_NATURAL_REGENERATION) &&
			(tickRegeneration(PlayerConfig.Food.FAST_REGEN) || tickRegeneration(PlayerConfig.Food.SLOW_REGEN))
		) {
			// do nothing
		} else if (PlayerConfig.Food.ENABLE_STARVATION && foodLevel <= 0) {
			if (++tickTimer >= PlayerConfig.Food.STARVATION_TICKS) {
				tickTimer = 0

				val threshold = when (player.level().difficulty) {
					Difficulty.PEACEFUL -> Float.POSITIVE_INFINITY
					Difficulty.EASY -> PlayerConfig.Food.STARVATION_HEALTH_LIMIT_EASY.toFloat()
					Difficulty.NORMAL -> PlayerConfig.Food.STARVATION_HEALTH_LIMIT_NORMAL.toFloat()
					Difficulty.HARD -> PlayerConfig.Food.STARVATION_HEALTH_LIMIT_HARD.toFloat()
				}

				if (player.health > threshold) {
					player.hurt(player.damageSources().starve(), 1.0f)
				}
			}
		} else {
			tickTimer = 0
		}
	}

	companion object {
		const val EXHAUSTION_PER_HUNGER_POINT = 4f
	}
}
